def factorial(x: float) -> float:
    """
    Return the factorial of x where x ∉ {-1, -2, -3, ...}.
    """
    from .exp import exp
    from .integrate import integrate
    if int(x) == x:
        if x < 0:
            raise ValueError("Factorial of a negative integer is undefined.")
        else:
            ans = 1
            for i in range(1,int(x)+1):
                ans *= i
            return ans
    def integrand(t):
        return t**x * exp(-t)
    return integrate(integrand, 0, 100)


def quintic(a:float | complex,b: float | complex,c: float | complex,d: float | complex,e: float | complex, f: float | complex) -> list:
    """
    Return the solutions of the quintic equation (form of ax⁵ + bx⁴ + cx³ + dx² + ex + f = 0).
    """
    from .quartic import quartic
    from .derivative import derivative
    if a == 0:
        raise ValueError("Value of 'a' must be different to 0.")
    if f == 0:
        quartic(a,b,c,d,e)
        quintic.x1, quintic.x2, quintic.x3, quintic.x4, quintic.x5 = 0, quartic.x1, quartic.x2, quartic.x3, quartic.x4
    elif a+b+c+d+e+f == 0:
        A = a
        B = b+A
        C = c+B
        D = d+C
        E = -f
        quartic(A,B,C,D,E)
        quintic.x1, quintic.x2, quintic.x3, quintic.x4, quintic.x5 = 1, quartic.x1, quartic.x2, quartic.x3, quartic.x4
    elif -a+b-c+d-e+f == 0:
        A = a
        B = b-A
        C = c-B
        D = d-C
        E = f
        quartic(A,B,C,D,E)
        quintic.x1, quintic.x2, quintic.x3, quintic.x4, quintic.x5 = -1, quartic.x1, quartic.x2, quartic.x3, quartic.x4
    else:
        quintic.x1 = 1
        def func(x):
            return a*x**5+b*x**4+c*x**3+d*x**2+e*x+f
        while True:
            g = func(quintic.x1)
            if abs(func(quintic.x1)) < 1e-12:
                break
            g_prime = derivative(func, quintic.x1)
            if abs(g_prime) < 1e-6:
                quintic.x1 += 1
            else:
                quintic.x1 -= g/g_prime
        A = a
        B = b+A*quintic.x1
        C = c+B*quintic.x1
        D = d+C*quintic.x1
        E = -f/quintic.x1
        quartic(A,B,C,D,E)
        quintic.x2, quintic.x3, quintic.x4, quintic.x5 = quartic.x1, quartic.x2, quartic.x3, quartic.x4
    solutions = [quintic.x1, quintic.x2, quintic.x3, quintic.x4, quintic.x5]
    result = []
    for solution in solutions:
        if solution not in result:
            result.append(solution)
    return result


def cbrt(x: float | complex) -> float | complex:
    """
    Return the cubic root of x.
    """
    from .sqrt import sqrt
    from .exp import exp
    from .atan import atan
    from .constants import pi
    if type(x) is complex:
        a = x.real
        b = x.imag
        if a == 0:
            return cbrt(b)*-1j
        if b == 0:
            return cbrt(a)
        if a > 0 and b > 0:
            return cbrt(sqrt(a**2+b**2))*exp(1j*(atan(b/a))/3)
        if a > 0 and b < 0:
            return cbrt(sqrt(a**2+b**2))*exp(1j*(2*pi-atan(-b/a))/3)
        if a < 0 and b > 0:
            return cbrt(sqrt(a**2+b**2))*exp(1j*(pi-atan(b/-a))/3)
        return cbrt(sqrt(a**2+b**2))*exp(1j*(pi+atan(b/a))/3)
    if x == 0:
        return 0.0
    y = x
    while True:
        f = y**3-x
        if abs(f) < 1e-12:
            return float(y)
        f_prime = 3*y**2
        if f_prime == 0:
            y += 1
        else:
            y -= f/f_prime


def ln(x: float | complex) -> float | complex:
    """
    Return the natural logarithm of x.
    """
    from .exp import exp
    from .atan import atan
    from .constants import pi
    if x == 0:
        raise ValueError("Value input must be different to 0.")
    if type(x) is not complex:
        if x > 0:
            if x == 1:
                return 0
            y = x
            while True:
                f = exp(y) - x
                f_prime = exp(y)
                if abs(f_prime) < 1e-6:
                    y += 1
                else:
                    y -= f/f_prime
                    if abs(f) < 1e-12:
                        return y
        y = abs(x)
        if abs(x) == 1:
            return pi*1j
        while True:
                f = exp(y) - abs(x)
                f_prime = exp(y)
                if abs(f_prime) < 1e-6:
                    y += 1
                else:
                    y -= f/f_prime
                    if abs(f) < 1e-12:
                        return y + pi*1j
    else:
        a = x.real
        b = x.imag
        if a == 0:
            return ln(b) + 1j*pi/2
        return ln(abs(x)) + 1j*atan(b/a)


def sin(x: float | complex, deg=False) -> float | complex:
    """
    Return the sine of x.

    Measured in radians as defualt.
    """
    from .constants import pi, e
    if deg:
        x = pi*(x/180)
    if type(x) is not complex:
        return ((e**(1j*x)-e**(1j*-x))/2j).real
    if abs(((e**(1j*x)-e**(1j*-x))/2j).imag) < 1e-6:
        return ((e**(1j*x)-e**(1j*-x))/2j).real
    return (e**(1j*x)-e**(1j*-x))/2j


def acos(x: float) -> float:
    """
    Return the arc cosine (measured in randians) of x.
    """
    from .asin import asin
    from .constants import pi
    if -1 <= x <= 1:
        return pi/2 - asin(x)
    else:
        raise ValueError("Value input must be in [-1 .. 1]")



def log(x: float, base=10) -> float | complex:
    """
    Return the logarithm of x in 'base'.

    Default base is 10 for decimal logarithm.
    """
    from .ln import ln
    return(ln(x)/ln(base))



def atan(x: float) -> float:
    """
    Return the arc tangent (measured in radians) of x.
    """
    from .integrate import integrate
    def integrand(t):
        return 1/(1+t**2)
    return integrate(integrand, 0, x)



def derivative(func: 'function', value: float | complex) -> float | complex:
    """
    Return the derivative of a function 'func' at a specific value.
    """
    h=1e-10
    ans=(func(value+h)-func(value))/h
    return ans


def integrate(integrand, lower_limit: float, upper_limit: float) -> float:
    """
    Return the result of the integral of 'integrand' from 'lower_limit' to 'upper_limit'.
    """
    n = 10000
    if upper_limit >= 5000:
        upper_limit = 5000
    if upper_limit >= 5000:
        upper_limit = -5000
    if lower_limit >= 5000:
        lower_limit = 5000
    if lower_limit >= 5000:
        lower_limit = -5000
    
    segment_width = (upper_limit - lower_limit) / n
    result = 0.5 * (integrand(lower_limit) + integrand(upper_limit))
    for i in range(1,n):
        x_i = lower_limit + i * segment_width
        result += integrand(x_i)
    result *= segment_width
    if result >= 1e10:
        return float('inf')
    elif result <= -1e10:
        return float('-inf')
    return result


def is_prime(x: int) -> bool:
    """
    Return True if x is prime. False otherwise.
    """
    from .floor import floor
    from .sqrt import sqrt
    if int(x) != x or x < 0:
        raise TypeError("Value input must be a positive integer.")
    if x == 0 or x == 1:
        return False
    if x == 2:
        return True
    for i in range(2, floor(sqrt(x))+1):
        if int(x) % i == 0:
            return False
    return True



def tan(x: float | complex, deg=False) -> float | complex:
    """
    Return the tangent of x.

    Measured in radians as defualt.
    """
    from .sin import sin
    from .cos import cos
    from .constants import pi
    if deg:
        x = pi*(x/180)
    return sin(x)/cos(x)
